@@ -1,5 +1,5 @@
 /*  dvdisaster: Additional error correction for optical media.
- *  Copyright (C) 2004,2005 Carsten Gnoerlich.
+ *  Copyright (C) 2004-2006 Carsten Gnoerlich.
  *  Project home page: http://www.dvdisaster.com
  *  Email: carsten@dvdisaster.com  -or-  cgnoerlich@fsfe.org
  *
@@ -52,389 +52,11 @@ void PrepareDeadSector()
     }
 }
 
-/*
- * See if we can handle the specified .ecc file
- */
-
-int VerifyVersion(EccHeader *eh, int fatal)
-{  
-   if(Closure->version < eh->neededVersion)
-   {  if(fatal)
-       Stop(_("\n* This .ecc file requires dvdisaster V%d.%d.%d or higher.\n"
-	      "* Please visit http://www.dvdisaster.com for an upgrade."),
-	    eh->neededVersion/10000,
-	    (eh->neededVersion%10000)/100,
-	    eh->neededVersion%100);
-
-      return 1;
-   }
-
-   return 0;
-}
 
 /*
- * Append the given file suffix if none is already present
+ * Create the error correction file
  */
 
-char *ApplyAutoSuffix(char *filename, char *suffix)
-{  char *out;
-
-   if(!filename || !*filename || strrchr(filename, '.')) 
-     return filename;
-
-   out = g_strdup_printf("%s.%s",filename,suffix);
-   g_free(filename);
-   
-   return out;
-}
-
-/***
- *** Open the image and ecc files and fill in the respective Info structs.
- ***/
-
-void OpenImageAndEcc(ImageInfo **iptr, EccInfo **eptr, int mode)
-{  ImageInfo *ii = NULL;
-   EccInfo   *ei = NULL; 
-
-   /*** Open the ecc file  */
-
-   if(eptr)
-   {
-      *eptr = ei = g_malloc0(sizeof(EccInfo));
-      ei->eh = g_malloc0(sizeof(EccHeader));
-
-      if(!(ei->file = fopen(Closure->eccName, mode & WRITEABLE_ECC ? "w" : "r")))
-      {  FreeEccInfo(ei);
-         *eptr = NULL;
-	 ei = NULL;
-
-	 if(mode & PRINT_MODE)  /* missing ecc file no problem in print mode */
-	 {  goto open_image;
-	 }
-         else  Stop(_("Can't open %s:\n%s"),Closure->eccName,strerror(errno));
-      }
-
-#ifdef SYS_MINGW
-      setmode(fileno(ei->file), O_BINARY);
-#endif
-
-      if(!(mode & WRITEABLE_ECC))
-      {   int n = fread(ei->eh, sizeof(EccHeader), 1, ei->file);
-
-          if(n != 1)
-	    Stop(_("Can't read ecc header:\n%s"),strerror(errno));
-
-	  /*** See if we can use the ecc file */
-
-	  if(   ei->eh->method_flags[0] != 1 || ei->eh->method_flags[1] 
-	     || ei->eh->method_flags[2] || ei->eh->method_flags[3])
-	    Stop(_("\nEcc file has been created with version 0.40.7.\n"
-		   "Please use version 0.41.1 or higher to create the .ecc files.\n"
-		   "I pledge not to change the file format again until version 2.0 :-)\n"));
-
-	  VerifyVersion(ei->eh,1);
-
-	  ei->sectors = uchar_to_gint64(ei->eh->sectors);
-	  rewind(ei->file);
-      }
-   }
-
-   /*** Open the image file */
-
-open_image:
-
-   if(iptr)
-   {  gint64 img_size;
-      int n,file_flags;
-      gint64 fp_sector;
-      unsigned char buf[2048];
-      struct MD5Context md5ctxt;
-
-      *iptr = ii = g_malloc0(sizeof(ImageInfo));
-
-      if(!(mode & PRINT_MODE))
-	PrintLog(_("\nOpening %s"), Closure->imageName);
-
-      if(!LargeStat(Closure->imageName, &img_size))
-      {  if(mode & PRINT_MODE)  /* no error in print mode */ 
-	 {  FreeImageInfo(ii);
- 	    *iptr = NULL;
-	    return;
-	 }
-
-         PrintLog(_(": not present.\n"));
-         g_free(ii); *iptr=NULL;
-	 Stop(_("Image file %s not present."),Closure->imageName);
-      }
-
-      file_flags = mode & WRITEABLE_IMAGE ? O_RDWR : O_RDONLY;
-
-      if(!(ii->file = LargeOpen(Closure->imageName, file_flags, IMG_PERMS)))
-      {  if(Closure->guiMode)
-	 {  g_free(ii); *iptr=NULL;
-	    PrintLog(": %s", strerror(errno));
-	    Stop(_("Can't open %s:\n%s"),Closure->imageName,strerror(errno));
-	 }
-         else Stop(": %s", strerror(errno));
-      }
-
-      ii->size = img_size;
-      CalcSectors(img_size, &ii->sectors, &ii->inLast);
-
-      if(!(mode & PRINT_MODE))
-	PrintLog(_(": %lld medium sectors.\n"), ii->sectors);
-
-
-      /*** Calculate md5sum of the footprint sector.
-	   Use sector specified by .ecc file if possible, 
-	   else use built-in default  */
-
-      if(!ei || (mode & WRITEABLE_ECC))
-	   fp_sector = FOOTPRINT_SECTOR;
-      else fp_sector = ei->eh->fpSector;
-
-      LargeSeek(ii->file, fp_sector*2048);
-
-      MD5Init(&md5ctxt);
-      n = LargeRead(ii->file, buf, 2048);
-      if(n != 2048)
-	Stop(_("could not read image sector %lld (only %d bytes):\n%s"),
-	     fp_sector,n,strerror(errno));
-      MD5Update(&md5ctxt, buf, n);
-      MD5Final(ii->mediumFP, &md5ctxt);
-      if(!memcmp(buf,Closure->deadSector,2048))  /* No sector, no md5sum */
-	memset(ii->mediumFP, 0 ,16);
-      LargeSeek(ii->file, 0);                     /* rewind */
-   }
-}
-
-/*
- * Scan the image for missing blocks.
- * If the ecc file is present, also compare the CRC sums.
- * If CREATE_CRC is requested, calculate the CRC sums.
- */
-
-#define CRCBUFSIZE (1024*256)
-
-void ScanImage(ImageInfo *ii, EccInfo *ei, int mode, GtkWidget *pbar)
-{  EccHeader eh; 
-   unsigned char buf[2048];
-   guint32 crcbuf[CRCBUFSIZE];
-   int crcidx = 0;
-   struct MD5Context image_md5;
-   gint64 eh_sectors = 0;
-   gint64 s, first_missing, last_missing;
-   gint64 prev_missing = 0;
-   gint64 prev_crc_errors = 0;
-   int last_percent,current_missing;
-   char *msg;
-
-   /* Read the ecc file header */
-
-   if(ei && mode != CREATE_CRC)
-   {   rewind(ei->file);
-       fread(&eh, sizeof(EccHeader), 1, ei->file);
-       eh_sectors = uchar_to_gint64(eh.sectors);
-   }     
-
-   /* Position behind the ecc file header,
-      initialize CRC buffer pointers */
-
-   if(ei)
-   {  if(fseeko(ei->file, (gint64)sizeof(EccHeader), SEEK_SET) == -1)
-         Stop(_("fseek failed skipping ecc header: %s"),strerror(errno));
-
-      crcidx = (mode & CREATE_CRC) ? 0 : CRCBUFSIZE;
-      MD5Init(&ei->md5Ctxt);            /* md5sum of CRC portion of ecc file */
-   }
-
-   /* Prepare for scanning the image and calculating its md5sum */
-
-   MD5Init(&image_md5);              /* md5sum of image file itself */
-   LargeSeek(ii->file, 0);            /* rewind image file */   
-      
-   if(mode & PRINT_MODE)
-        msg = _("- testing sectors  : %3d%%");
-   else msg = _("Scanning image sectors: %3d%%");
-
-   last_percent = 0;
-   ii->sectorsMissing = 0;
-   first_missing = last_missing = -1;
-
-   /* Go through all sectors and look for the "dead sector marker" */
-   
-   for(s=0; s<ii->sectors; s++)
-   {  int n,percent;
-
-      /* Check for user interruption */
-
-      if(Closure->stopActions)   
-      {  ii->sectorsMissing += ii->sectors - s;
-         return;
-      }
-
-      /* Read the next sector */
-
-      n = LargeRead(ii->file, buf, 2048);
-      if(n != 2048 && (s != ii->sectors - 1 || n != ii->inLast))
-	Stop(_("premature end in image (only %d bytes): %s\n"),n,strerror(errno));
-
-      /* Look for the dead sector marker */
-
-      current_missing = !memcmp(buf, Closure->deadSector, n);
-      if(current_missing)
-      {  if(first_missing < 0) first_missing = s;
-         last_missing = s;
-	 ii->sectorsMissing++;
-      }
-
-      /* Report dead sectors. Combine subsequent missing sectors into one report. */
-
-      if(mode & PRINT_MODE)
-	if(!current_missing || s==ii->sectors-1)
-	{  if(first_missing>=0)
-	    {   if(first_missing == last_missing)
-		     PrintCLI(_("* missing sector   : %lld\n"), first_missing);
-		else PrintCLI(_("* missing sectors  : %lld - %lld\n"), first_missing, last_missing);
-	      first_missing = -1;
-	   }
-	}
-
-      if(ei)   /* Do something with the CRC portion of the .ecc file */
-      {
-	 /* If creation of the CRC32 is requested, do that. */
-
-	 if(mode & CREATE_CRC)
-	 {  crcbuf[crcidx++] = Crc32(buf, 2048);
-
-	    if(crcidx >= CRCBUFSIZE)  /* write out CRC buffer contents */
-	    {  size_t size = CRCBUFSIZE*sizeof(guint32);
-
-	       MD5Update(&ei->md5Ctxt, (unsigned char*)crcbuf, size);
-	       if(fwrite(crcbuf, 1, size, ei->file) != size)
-		 Stop(_("Error writing CRC information: %s"),strerror(errno));
-
-	       crcidx = 0;
-	    }
-	 }
-
-	 /* else do the CRC32 check. Missing sectors are skipped in the CRC report. */
-	 
-	 else if(s < eh_sectors)
-	 {  guint32 crc = Crc32(buf, 2048); 
-
-            /* If the CRC buf is exhausted, refill. */
-
-            if(crcidx >= CRCBUFSIZE)
-	    {  size_t remain = ii->sectors-s;
-	       size_t size;
-
-	       if(remain < CRCBUFSIZE)
-		    size = remain*sizeof(guint32);
-	       else size = CRCBUFSIZE*sizeof(guint32);
-
-	       if(fread(crcbuf, 1, size, ei->file) != size)
-		 Stop(_("Error reading CRC information: %s"),strerror(errno));
-
-	       crcidx = 0;
-	    }
-
-	    if(crc != crcbuf[crcidx++] && !current_missing)
-	    {  PrintCLI(_("* CRC error, sector: %lld\n"), s);
-	       ii->crcErrors++;
-	    }
-	 }
-      }
-
-      MD5Update(&image_md5, buf, n);  /* update image md5sum */
-
-      if(Closure->guiMode && mode & PRINT_MODE) 
-	    percent = (COMPARE_IMAGE_SEGMENTS*s)/ii->sectors;
-      else  percent = (100*s)/ii->sectors;
-      if(last_percent != percent) 
-      {  PrintProgress(msg,percent);
-
-         if(pbar)
-	   SetProgress(pbar, percent, 100);
-
-	 if(Closure->guiMode && mode & PRINT_MODE)
-	 {  AddCompareValues(percent, ii->sectorsMissing, ii->crcErrors,
-			     ii->sectorsMissing - prev_missing,
-			     ii->crcErrors - prev_crc_errors);
-	    prev_missing = ii->sectorsMissing;
-	    prev_crc_errors = ii->crcErrors;
-	 }
-
-	 last_percent = percent;
-      }
-   }
-
-   /*** Flush the rest of the CRC buffer */
-
-   if((mode & CREATE_CRC) && crcidx)
-   {  size_t size = crcidx*sizeof(guint32);
-
-      MD5Update(&ei->md5Ctxt, (unsigned char*)crcbuf, size);
-      if(fwrite(crcbuf, 1, size, ei->file) != size)
-	Stop(_("Error writing CRC information: %s"),strerror(errno));
-   }
-
-   /*** The image md5sum can only be calculated if all blocks have been successfully read. */
-
-   MD5Final(ii->mediumSum, &image_md5);
-   PrintProgress(msg,100);
-
-   LargeSeek(ii->file, 0);
-}
-
-/*
- * Free the ImageInfo and EccInfo stuff
- */
-
-void FreeImageInfo(ImageInfo *ii)
-{
-   if(ii->file) 
-     if(!LargeClose(ii->file))
-       Stop(_("Error closing image file:\n%s"), strerror(errno));
-
-   g_free(ii);
-}
-
-void FreeEccInfo(EccInfo *ei)
-{
-   if(ei->file) 
-     if(fclose(ei->file))
-       Stop(_("Error closing error correction file:\n%s"), strerror(errno));
-
-   if(ei->eh) g_free(ei->eh);
-   g_free(ei);
-}
-
-/*
- * Create ecc information for the medium
- */
-
-/* Remove the image file if requested */
-
-void UnlinkImage(GtkWidget *label)
-{
-   if(LargeUnlink(Closure->imageName))
-   {    PrintLog(_("\nImage file %s deleted.\n"),Closure->imageName);
-
-        if(Closure->guiMode)
-	  SetLabelText(GTK_LABEL(label),
-		       _("\nImage file %s deleted.\n"), Closure->imageName);
-   }
-   else 
-   {  if(!Closure->guiMode)
-       PrintLog("\n");
-
-       ModalWarning(GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, NULL,
-		    _("Image file %s not deleted: %s\n"),
-		    Closure->imageName, strerror(errno));
-   }
-}
-
 void CreateEcc(void)
 {  Method *method = FindMethod(Closure->methodName); 
 
@@ -453,27 +75,37 @@ void CreateEcc(void)
  */
 
 void FixEcc(void)
-{  EccInfo *ei;
-   Method *method; 
-   char method_name[5];
-
-   /*** Peek into Ecc header to determine the method */
-
-   OpenImageAndEcc(NULL, &ei, READABLE_ECC);
-   memcpy(method_name, ei->eh->method, 4); method_name[4] = 0;
-   FreeEccInfo(ei);
-
-   /*** GUI mode has already verified that the method exists before
-	calling us so the following Stop() will never execute in GUI mode. */
-
-   method = FindMethod(method_name);
-   if(!method) Stop(_("\nUnknown method %s.\n"),method_name);
-
-   /*** Dispatch to proper method */
+{  Method *method; 
    
+   /* Error handling is done within EccFileMethod() */
+
+   method = EccFileMethod(TRUE);
+ 
+   /* Dispatch to the proper method */
+  
    method->fix(method);
 }
 
+/*
+ * Verfiy the image against ecc data 
+ */
+
+void Verify(void)
+{  Method *method; 
+   
+  /* If something is wrong with the .iso or .ecc files
+     we fall back to the RS01 method for comparing 
+     since it is robust against missing files. */
+
+  if(!(method = EccFileMethod(FALSE)))
+    if(!(method = FindMethod("RS01")))
+      Stop(_("RS01 method not available for comparing files."));
+ 
+   /* Dispatch to the proper method */
+
+   method->verify(method);
+}
+
 /*
  * The all-famous main() loop 
  */
@@ -484,7 +116,7 @@ typedef enum
    MODE_CREATE,
    MODE_HELP,
    MODE_FIX,
-   MODE_PRINT, 
+   MODE_VERIFY, 
    MODE_READ, 
    MODE_SCAN,
    MODE_SEQUENCE, 
@@ -508,12 +140,14 @@ typedef enum
    MODIFIER_DEBUG,
    MODIFIER_FILL_UNREADABLE,
    MODIFIER_KEEP_STYLE,
-   MODIFIER_PARSE_UDF,
+   MODIFIER_QUERY_SIZE,
    MODIFIER_RANDOM_SEED,
+   MODIFIER_SIMULATE_DEFECTS,
    MODIFIER_SPEED_WARNING, 
    MODIFIER_SPINUP_DELAY, 
    MODIFIER_SPLIT_FILES,
-   MODIFIER_TRUNCATE
+   MODIFIER_TRUNCATE,
+   MODIFIER_VERSION
 } run_mode;
 
 int main(int argc, char *argv[])
@@ -523,13 +157,27 @@ int main(int argc, char *argv[])
    char *default_device;
    char *read_range = NULL;
 #ifdef WITH_NLS_YES
-   char *locale_test,src_locale_path[strlen(SRCDIR)+10];
+   char *locale_test;
+ #ifndef SYS_MINGW
+   char src_locale_path[strlen(SRCDIR)+10];
+ #else
+   char *bin_locale_path = NULL;
+ #endif
 #endif
 
 #ifdef WITH_MEMDEBUG_YES
     atexit(check_memleaks);
 #endif
 
+#ifdef SYS_MINGW
+    /*** Create a named mutex so that the installer can detect
+	 that we are running. A malicious user may cause this 
+	 to fail be reserving the mutex for himself, 
+	 so we do not care if the mutex can not be created. */
+
+    CreateMutex(NULL, FALSE, "dvdisaster");
+#endif
+
     /*** Setup the global closure. */
 
     InitClosure();
@@ -550,21 +198,39 @@ int main(int argc, char *argv[])
 
        lang_id = GetUserDefaultLangID(); 
        switch(PRIMARYLANGID(lang_id))
-       {  case 0x07:  /* German */
+       {  case LANG_CZECH:
+	     g_setenv("LANG", "cs_CZ", 1);
+#ifdef WIN_CONSOLE
+	     g_setenv("OUTPUT_CHARSET", "CP852", 1);
+#else
+	     g_setenv("OUTPUT_CHARSET", "CP1250", 1);
+#endif
+	     break;
+
+          case LANG_GERMAN:
 	     g_setenv("LANG", "de_DE", 1);
 #ifdef WIN_CONSOLE
 	     g_setenv("OUTPUT_CHARSET", "CP850", 1);
 #else
-	     g_setenv("OUTPUT_CHARSET", "iso-8859-1", 1);
+	     g_setenv("OUTPUT_CHARSET", "CP1252", 1);
 #endif
 	     break;
 
-          case 0x10:  /* Italian */
+          case LANG_ITALIAN:
 	     g_setenv("LANG", "it_IT", 1);
 #ifdef WIN_CONSOLE
 	     g_setenv("OUTPUT_CHARSET", "CP850", 1);
 #else
-	     g_setenv("OUTPUT_CHARSET", "iso-8859-1", 1);
+	     g_setenv("OUTPUT_CHARSET", "CP1252", 1);
+#endif
+	     break;
+
+          case LANG_SWEDISH:
+	     g_setenv("LANG", "sv_SV", 1);
+#ifdef WIN_CONSOLE
+	     g_setenv("OUTPUT_CHARSET", "CP850", 1);
+#else
+	     g_setenv("OUTPUT_CHARSET", "CP1252", 1);
 #endif
 	     break;
        }
@@ -575,11 +241,12 @@ int main(int argc, char *argv[])
     setlocale(LC_MESSAGES, "");
     textdomain("dvdisaster");
 
+#ifndef SYS_MINGW
     /* Try local source directory first */
 
     g_sprintf(src_locale_path,"%s/locale",SRCDIR);
     bindtextdomain("dvdisaster", src_locale_path);
-
+//printf("testing src %s\n", src_locale_path);
     /* TRANSLATORS: 
        This is a dummy entry which is supposed to translate into "ok".
        Please do not return anything else here. */
@@ -594,26 +261,29 @@ int main(int argc, char *argv[])
        if(getcwd(buf, 256))
        {  char locale_path[strlen(buf)+20];
           g_sprintf(locale_path,"%s/locale", buf);
+//printf("testing cwd %s\n", locale_path);
           bindtextdomain("dvdisaster", locale_path);
 	  locale_test = gettext("test phrase for verifying the locale installation");
        }
     }
+#endif
 
 #ifdef SYS_MINGW
     /* Try the directory where our executable comes from.
-       Only possible under Windows. */
+       This only possible under Windows, and should cover all cases. */
 
-    if(strcmp(locale_test, "ok"))
-    {  char locale_path[strlen(Closure->binDir)+20];
-
-       g_sprintf(locale_path,"%s/locale", Closure->binDir);
-       bindtextdomain("dvdisaster", locale_path);
-       locale_test = gettext("test phrase for verifying the locale installation");
-    }
+    bin_locale_path = g_strdup_printf("%s\\locale", Closure->binDir);
+//printf("testing bin %s\n", bin_locale_path);
+    bindtextdomain("dvdisaster", bin_locale_path);
+    locale_test = gettext("test phrase for verifying the locale installation");
+    g_free(bin_locale_path);
 #endif
 
     /* Last resort: fall back to global locale */
 
+//    if(strcmp(locale_test, "ok"))
+//      printf("fallback used\n");
+
     if(strcmp(locale_test, "ok"))
       bindtextdomain("dvdisaster", LOCALEDIR);  
 #endif /* WITH_NLS_YES */
@@ -661,10 +331,10 @@ int main(int argc, char *argv[])
 #ifdef SYS_MINGW
 	{"list", 0, 0, 'l' },
 #endif
-	{"method", 1, 0, 'm' },
-	{"parse-udf", 0, 0, MODIFIER_PARSE_UDF },
+	{"method", 2, 0, 'm' },
+	{"query-size", 1, 0, MODIFIER_QUERY_SIZE },
         {"prefix", 1, 0, 'p'},
-	{"random-errors", 0, 0, MODE_RANDOM_ERR },
+	{"random-errors", 1, 0, MODE_RANDOM_ERR },
 	{"random-image", 1, 0, MODE_RANDOM_IMAGE },
 	{"random-seed", 1, 0, MODIFIER_RANDOM_SEED },
 	{"read", 2, 0,'r'},
@@ -672,13 +342,15 @@ int main(int argc, char *argv[])
 	{"scan", 2, 0,'s'},
 	{"send-cdb", 1, 0, MODE_SEND_CDB},
 	{"show-sector", 1, 0, MODE_SHOW_SECTOR},
+	{"sim-defects", 1, 0, MODIFIER_SIMULATE_DEFECTS},
 	{"speed-warning", 2, 0, MODIFIER_SPEED_WARNING},
 	{"spinup-delay", 1, 0, MODIFIER_SPINUP_DELAY},
 	{"split-files", 0, 0, MODIFIER_SPLIT_FILES},
 	{"test", 0, 0, 't'},
 	{"truncate", 2, 0, MODIFIER_TRUNCATE},
 	{"unlink", 0, 0, 'u'},
-       	{"version", 0, 0, 'v'},
+       	{"verbose", 0, 0, 'v'},
+	{"version", 0, 0, MODIFIER_VERSION},
 	{"zero-unreadable", 0, 0, MODE_ZERO_UNREADABLE},
         {0, 0, 0, 0}
       };
@@ -699,7 +371,7 @@ int main(int argc, char *argv[])
 		   break;
          case 'c': mode = MODE_SEQUENCE; sequence |= 1<<MODE_CREATE; break;
          case 'f': mode = MODE_SEQUENCE; sequence |= 1<<MODE_FIX; break;
-         case 't': mode = MODE_SEQUENCE; sequence |= 1<<MODE_PRINT; break;
+         case 't': mode = MODE_SEQUENCE; sequence |= 1<<MODE_VERIFY; break;
          case 'u': Closure->unlinkImage = TRUE; break;
          case 'h': mode = MODE_HELP; break;
          case 'i': if(optarg) 
@@ -717,17 +389,23 @@ int main(int argc, char *argv[])
                    }
 	           else {  ListMethods(); FreeClosure(); exit(EXIT_SUCCESS); }
 	           break;
-         case 'n': if(optarg) Closure->redundancy = g_strdup(optarg); break;
+         case 'n': if(optarg) 
+		   {  Closure->redundancy = g_strdup(optarg); 
+		      Closure->mediumSize = (gint64)atoll(optarg);
+		      break;
+		   }
          case 'e': if(optarg) 
 	           {  g_free(Closure->eccName);
 		      Closure->eccName = g_strdup(optarg);
 		   }
 	           break;
          case 'p': if(optarg) 
-	           {  Closure->eccName = g_malloc(strlen(optarg)+5);
+		   {  g_free(Closure->imageName);
+		      g_free(Closure->eccName);
+		      Closure->eccName = g_malloc(strlen(optarg)+5);
 		      Closure->imageName = g_malloc(strlen(optarg)+5);
 		      g_sprintf(Closure->eccName,"%s.ecc",optarg);
-		      g_sprintf(Closure->imageName,"%s.img",optarg);
+		      g_sprintf(Closure->imageName,"%s.iso",optarg);
 		   }
 	           break;
          case 'd': if(optarg) 
@@ -735,9 +413,7 @@ int main(int argc, char *argv[])
 	              Closure->device = g_strdup(optarg); 
 	              break;
                    }
-         case 'v': PrintCLI(_("\ndvdisaster version %s\n\n"), VERSION);
-	 	   FreeClosure();
-	           exit(EXIT_SUCCESS); 
+         case 'v': Closure->verbose = TRUE;
 	           break;
 
          case  0 : break; /* flag argument */
@@ -774,12 +450,19 @@ int main(int argc, char *argv[])
          case MODIFIER_DEBUG:
 	   Closure->debugMode = TRUE;
 	   break;
-         case MODIFIER_PARSE_UDF:
-	   Closure->parseUDF = TRUE;
+         case MODIFIER_QUERY_SIZE:
+	        if(!strcmp(optarg, "drive")) Closure->querySize = 0;
+	   else if(!strcmp(optarg, "udf"))   Closure->querySize = 1;
+	   else if(!strcmp(optarg, "ecc"))   Closure->querySize = 2;
+           else Stop("--query-size requires one of these arguments: drive udf ecc\n");
 	   break;
          case MODIFIER_RANDOM_SEED:
 	   if(optarg) Closure->randomSeed = atoi(optarg);
 	   break;
+         case MODIFIER_SIMULATE_DEFECTS:
+	   if(optarg) Closure->simulateDefects = atoi(optarg);
+	   else Closure->simulateDefects = 10;
+	   break;
          case MODIFIER_SPINUP_DELAY:
 	   if(optarg) Closure->spinupDelay = atoi(optarg);
 	   break;
@@ -796,6 +479,11 @@ int main(int argc, char *argv[])
          case MODIFIER_CAV_SPEED:
 	   Closure->driveSpeed = -atoi(optarg);
 	   break;
+         case MODIFIER_VERSION:
+	   PrintCLI(_("\ndvdisaster version %s\n\n"), VERSION);
+	   FreeClosure();
+	   exit(EXIT_SUCCESS); 
+	   break;
          case MODE_SEND_CDB: 
 	   mode = MODE_SEND_CDB;
 	   debug_arg = g_strdup(optarg);
@@ -806,6 +494,7 @@ int main(int argc, char *argv[])
 	   break;
          case MODE_RANDOM_ERR:
 	   mode = MODE_RANDOM_ERR;
+	   debug_arg = g_strdup(optarg);
 	   break;
          case MODE_RANDOM_IMAGE:
 	   mode = MODE_RANDOM_IMAGE;
@@ -868,7 +557,7 @@ int main(int argc, char *argv[])
 
    if(Closure->autoSuffix)
    {  Closure->eccName   = ApplyAutoSuffix(Closure->eccName, "ecc");
-      Closure->imageName = ApplyAutoSuffix(Closure->imageName, "img");
+      Closure->imageName = ApplyAutoSuffix(Closure->imageName, "iso");
    }
 
    /*** Dispatch action depending on mode.
@@ -881,7 +570,9 @@ int main(int argc, char *argv[])
 	  ReadMediumLinear((gpointer)1);
 
 	if(sequence & 1<<MODE_READ)
-	{  if(Closure->adaptiveRead) 
+	{  if(sequence & 1<<MODE_CREATE) 
+	      Closure->readAndCreate = TRUE;
+	   if(Closure->adaptiveRead) 
 	        ReadMediumAdaptive((gpointer)0);
 	   else ReadMediumLinear((gpointer)0);
 	}
@@ -892,8 +583,8 @@ int main(int argc, char *argv[])
 	if(sequence & 1<<MODE_FIX)
 	  FixEcc();
 
-	if(sequence & 1<<MODE_PRINT)
-	  Compare();
+	if(sequence & 1<<MODE_VERIFY)
+	  Verify();
 	break;
 
       case MODE_BYTESET:
@@ -913,7 +604,7 @@ int main(int argc, char *argv[])
 	 break;
 
       case MODE_RANDOM_ERR:
-	 RandomError(Closure->imageName);
+	 RandomError(Closure->imageName, debug_arg);
 	 break;
 
       case MODE_RANDOM_IMAGE:
@@ -943,7 +634,6 @@ int main(int argc, char *argv[])
 
    if(mode == MODE_HELP)
    {  
-#ifndef QUIET
      /* TRANSLATORS: Program options like -r and --read are not to be translated
 	to avoid confusion when discussing the program in international forums. */
       PrintCLI(_("\nCommon usage examples:\n"
@@ -952,13 +642,13 @@ int main(int argc, char *argv[])
 	     "  dvdisaster -c,--create # Create .ecc information for the medium image.\n"
 	     "  dvdisaster -f,--fix    # Try to fix medium image using .ecc information.\n"
 	     "  dvdisaster -s,--scan   # Scan the medium for read errors.\n"
-	     "  dvdisaster -t,--test   # Test integrity of the .img and .ecc files.\n"
-	     "  dvdisaster -u,--unlink # Delete .img files (when other actions complete)\n\n"));
+	     "  dvdisaster -t,--test   # Test integrity of the .iso and .ecc files.\n"
+	     "  dvdisaster -u,--unlink # Delete .iso files (when other actions complete)\n\n"));
 
       PrintCLI(_("Drive and file specification:\n"
 	     "  -d,--device device     - read from given device   (default: %s)\n"
-	     "  -p,--prefix prefix     - prefix of .img/.ecc file (default: medium.*  )\n"
-	     "  -i,--image  imagefile  - name of image file       (default: medium.img)\n"
+	     "  -p,--prefix prefix     - prefix of .iso/.ecc file (default: medium.*  )\n"
+	     "  -i,--image  imagefile  - name of image file       (default: medium.iso)\n"
 	     "  -e,--ecc    eccfile    - name of parity file      (default: medium.ecc)\n"),
 	     Closure->device);
 
@@ -970,13 +660,16 @@ int main(int argc, char *argv[])
 
       PrintCLI(_("Tweaking options (see manual before using!)\n"
 	     "  -j,--jump n            - jump n sectors forward after a read error (default: 16)\n"
-	     "  -n,--redundancy n%%     - sets redundancy for error correction (default: 14.3%%)\n"
+	     "  -m n                   - list/select error correction methods (default: RS01)\n" 
+	     "  -n,--redundancy n%%     - error correction file redundancy (in percent), or\n"
+	     "                           maximum error correction image size (in sectors)\n"
+	     "  -v,--verbose           - more diagnostic messages\n"
 	     "  --adaptive-read        - use optimized strategy for reading damaged media\n"
-	     "  --auto-suffix          - automatically add .img and .ecc file suffixes\n"
+	     "  --auto-suffix          - automatically add .iso and .ecc file suffixes\n"
 	     "  --cache-size n         - image cache size in MB during -c mode (default: 32MB)\n"
 	     "  --dao                  - assume DAO disc; do not trim image end\n"
 	     "  --fill-unreadable n    - fill unreadable sectors with byte n\n"
-	     "  --parse-udf            - use information from ISO/UDF filesystem\n"
+      	     "  --query-size n         - query drive/udf/ecc for image size (default: ecc)\n"         
 	     "  --speed-warning n      - print warning if speed changes by more than n percent\n"
 	     "  --spinup-delay n       - wait n seconds for drive to spin up\n"
 	     "  --split-files          - split image into files <= 2GB\n\n"));
@@ -985,20 +678,20 @@ int main(int argc, char *argv[])
 	     "  --keep-style           - do not override the style settings\n\n"));
 
       if(Closure->debugMode)
-	PrintCLI(_("Debugging options (purposefully undocumented and possibly harmful)\n"
+      { PrintCLI(_("Debugging options (purposefully undocumented and possibly harmful)\n"
 	     "  --debug           - enables the following options\n"
 	     "  --byteset s,i,b   - set byte i in sector s to b\n"
 	     "  --erase sector    - erase the given sector\n"
 	     "  --erase n-m       - erase sectors n - m, inclusively\n"
-	     "  --random-errors   - seed image with (correctable) random errors\n"
-	     "  --random-image n  - create image with n sectors or random numbers\n"
-	     "  --random-seed n   - random seed for above function\n"
- 	     "  --send-cdb arg    - executes given cdb at drive; kills system if used wrong\n"  
+	     "  --random-errors r,e seed image with (correctable) random errors\n"
+	     "  --random-image n  - create image with n sectors of random numbers\n"
+	     "  --random-seed n   - random seed for built-in random number generator\n"
+	     "  --send-cdb arg    - executes given cdb at drive; kills system if used wrong\n"
 	     "  --show-sector n   - shows hexdump of the given sector\n"
+	     "  --sim-defects n   - simulate n%% defective sectors on medium\n"
 	     "  --truncate n      - truncates image to n sectors\n"
 	     "  --zero-unreadable - replace the \"unreadable sector\" markers with zeros\n\n"));
-#endif
-
+      }
       FreeClosure();
       exit(EXIT_FAILURE);
    }
