@@ -1,5 +1,5 @@
 /*  dvdisaster: Additional error correction for optical media.
- *  Copyright (C) 2004,2005 Carsten Gnoerlich.
+ *  Copyright (C) 2004-2006 Carsten Gnoerlich.
  *  Project home page: http://www.dvdisaster.com
  *  Email: carsten@dvdisaster.com  -or-  cgnoerlich@fsfe.org
  *
@@ -66,7 +66,6 @@
 #endif
 
 #include "md5.h"
-#include "version.h"
 
 #ifndef G_THREADS_ENABLED
  #error "need multithreading glib2"
@@ -87,6 +86,18 @@
 #define IMG_PERMS (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)
 #endif
 
+/* Using round() is preferred over rint() on systems which have it */
+
+#ifndef HAVE_ROUND
+ #define round(x) rint(x)
+#endif
+
+/* Some standard media sizes */
+
+#define CDR_SIZE         (351*1024)
+#define DVD_SL_SIZE      2295104  /* DVD+R/RW size used at least common denominator */
+#define DVD_DL_SIZE 	 4171712  /* also seen: 4148992 4173824  */
+
 /***
  *** Our global closure (encapsulation of global variables)
  ***/
@@ -94,6 +105,7 @@
 typedef struct _GlobalClosure
 {  int version;         /* Integer number representing current program version */
    char *cookedVersion; /* version string formatted for GUI use */
+   gint8 releaseFlags;  /* flags marking release status */
    char *device;        /* currently selected device to read from */
    GPtrArray *deviceNames;  /* List of drive names */
    GPtrArray *deviceNodes;  /* List of device nodes (C: or /dev/foo) */
@@ -103,6 +115,13 @@ typedef struct _GlobalClosure
    char *methodName;    /* Name of currently selected codec */
    gint64 readStart;    /* Range to read */
    gint64 readEnd;
+   gint64 cdSize;       /* Maximum cd size (for RS02 type images) */
+   gint64 dvdSize1;     /* Maximum 1-layer dvd size (for RS02 type images) */
+   gint64 dvdSize2;     /* Maximum 2-layer dvd size (for RS02 type images) */
+   gint64 savedCDSize;  /* Undo values for above */
+   gint64 savedDVDSize1;
+   gint64 savedDVDSize2;
+   gint64 mediumSize;   /* Maximum medium size (for RS02 type images) */
    int cacheMB;         /* Cache setting for the parity codec, in megabytes */
    int sectorSkip;      /* Number of sectors to skip after read error occurs */
    char *redundancy;    /* Error correction code redundancy */
@@ -114,15 +133,20 @@ typedef struct _GlobalClosure
    int noTruncate;      /* do not truncate image at the end */
    int unlinkImage;     /* delete image after ecc file creation */
    int driveSpeed;      /* currently unused */
-   int debugMode;       /* may activate additional messages */
+   int debugMode;       /* may activate additional features */
+   int verbose;         /* may activate additional messages */
    int splitFiles;      /* limit image files to 2GB */
-   int autoSuffix;      /* automatically extend files with suffices .img/.ecc */
-   int parseUDF;        /* use information from the ISO/UDF filesystems */
+   int autoSuffix;      /* automatically extend files with suffices .iso/.ecc */
+   int querySize;       /* what sources are used for image size queries */
    int readAndCreate;   /* automatically create .ecc file after reading an image */
    int enableCurveSwitch; /* TRUE in readAndCreateMode after reading is complete */
+   int welcomeMessage;  /* just print dvdisaster logo if FALSE */
+   int dotFileVersion;  /* version of dotfile */
+   int simulateDefects; /* if >0, this is the percentage of simulated media defects */
   
    char *deadSector;    /* Copy of our "dead sector" marker. */
    char *dotFile;       /* path to .dvdisaster file */
+   char *logFile;       /* path to logfile */
    char *binDir;        /* place where the binary resides */
    char *docDir;        /* place where our documentation resides */
    char *browser;       /* Name of preferred WEB browser */
@@ -146,6 +170,11 @@ typedef struct _GlobalClosure
    unsigned char *scratchBufBase;  /* scratch buffer for use with SCSI commands; */
    unsigned char *scratchBuf;      /* 32768 bytes aligned at a 4096 byte boundary */
 
+   guint32 *crcCache;              /* sectorwise CRC32 for last image read */
+   char    *crcImageName;          /* file name of cached image */
+   unsigned char md5Cache[16];     /* md5sum of last image read */
+   
+
    /*** GUI-related things */
 
    int guiMode;         /* TRUE if GUI is active */
@@ -191,10 +220,12 @@ typedef struct _GlobalClosure
    GdkColor  *background,*grid,*black,*white;
    GdkColor  *red,*yellow,*green,*darkgreen,*blue;
    gint      lastPercent;
-   gint      lastPercentPlotted;
+   gint      lastSegment;
+   gint      lastPlotted;
+   gint      lastPlottedY;
    gint      percent; 
 
-   /*** Widgets for the reading/scanning action */
+   /*** Widgets for the linear reading/scanning action */
 
    GtkWidget *readLinearHeadline;
    GtkWidget *readLinearDrawingArea;
@@ -205,6 +236,8 @@ typedef struct _GlobalClosure
    GtkWidget *readLinearErrors;
    GtkWidget *readLinearFootline;
    GtkWidget *readLinearFootlineBox;
+   int    checkCrc;               /* these are passed between threads and */
+   gint64 crcErrors, readErrors;  /* must therefore be global */
 
    /*** Widgets for the adaptive reading action */
 
@@ -213,33 +246,7 @@ typedef struct _GlobalClosure
    struct _Spiral *readAdaptiveSpiral;
    char *readAdaptiveSubtitle;
    char *readAdaptiveErrorMsg;
-
-   /*** Widgets for compare action */
-
-   GtkWidget *cmpHeadline;
-
-   GtkWidget *cmpImageNotebook;
-   GtkWidget *cmpImageSectors;
-   GtkWidget *cmpChkSumErrors;
-   GtkWidget *cmpMissingSectors;
-   GtkWidget *cmpImageMd5Sum;
-   GtkWidget *cmpImageResult;
-
-   GtkWidget *cmpEccNotebook;
-   GtkWidget *cmpEccEmptyMsg;
-   GtkWidget *cmpEccCreatedBy;
-   GtkWidget *cmpEccMethod;
-   GtkWidget *cmpEccRequires;
-   GtkWidget *cmpEccMediumSectors;
-   GtkWidget *cmpEccImgMd5Sum;
-   GtkWidget *cmpEccFingerprint;
-   GtkWidget *cmpEccBlocks;
-   GtkWidget *cmpEccMd5Sum;
-   GtkWidget *cmpEccResult;
-
-   GtkWidget *cmpDrawingArea;
-   struct _Spiral *cmpSpiral;
-   PangoLayout *cmpLayout;
+   int additionalSpiralColor;
 
 } GlobalClosure;
 
@@ -261,6 +268,7 @@ typedef struct _LargeFile
    char *basename;
    char *suffix;
    int namelen;
+   gint64 size;
 } LargeFile;
 
 /***
@@ -285,7 +293,7 @@ typedef struct _ImageInfo
  */
 
 typedef struct _EccInfo
-{  FILE *file;                  /* file handle for ecc file */
+{  LargeFile *file;             /* file handle for ecc file */
    struct _EccHeader *eh;       /* the header struct below */
    gint64 sectors;              /* gint64 version of eh->sectors */ 
    struct MD5Context md5Ctxt;   /* md5 context of crc portion of file */
@@ -295,13 +303,20 @@ typedef struct _EccInfo
  *** The .ecc file header
  ***/
 
-#define FOOTPRINT_SECTOR 257   /* Sector currently used to calculate the footprint */
+#define FINGERPRINT_SECTOR 16 /* Sector currently used to calculate the fingerprint. */
+                              /* This is the ISO filesystem root sector which contains */
+                              /* the volume label and creation time stamps. */
+                              /* Versions upto 0.64 used sector 257, */
+                              /* but that was not a wise choice for CD media.*/
+
+#define MFLAG_DEVEL (1<<0)    /* for methodFlags[3] */
+#define MFLAG_RC    (1<<1)                      
 
 typedef struct _EccHeader
 {  gint8 cookie[12];           /* "*dvdisaster*" */
    gint8 method[4];            /* e.g. "RS01" */
-   gint8 method_flags[4];      /* for free use by the respective methods */
-   guint8 mediumFP[16];        /* fingerprint of first medium sector */ 
+   gint8 methodFlags[4];       /* 0-2 for free use by the respective methods; 3 see above */
+   guint8 mediumFP[16];        /* fingerprint of FINGERPRINT SECTOR */ 
    guint8 mediumSum[16];       /* complete md5sum of whole medium */
    guint8 eccSum[16];          /* md5sum of ecc code section of .ecc file */
    guint8 sectors[8];          /* number of sectors medium is supposed to have */
@@ -310,63 +325,51 @@ typedef struct _EccHeader
    gint32 creatorVersion;      /* which dvdisaster version created this */
    gint32 neededVersion;       /* oldest version which can decode this file */
    gint32 fpSector;            /* sector used to calculate mediumFP */
-   gint8 padding[4000];        /* pad to 4096 bytes: room for future expansion */
+   guint32 selfCRC;            /* CRC32 of EccHeader (currently RS02 only) -- since V0.66 --*/
+   guint8 crcSum[16];          /* md5sum of crc code section of RS02 .iso file  */
+   gint32 inLast;              /* bytes contained in last sector */
+   gint8 padding[3976];        /* pad to 4096 bytes: room for future expansion */
+
+  /* Note: Bytes 2048 and up are currently used by the RS02 codec
+           for a copy of the first ecc blocks CRC sums. */
 } EccHeader;
 
 /***
  *** dvdisaster.c
  ***/
 
-#define OPEN_READABLE     0
-#define OPEN_WRITEABLE    (1<<0)
-#define OPEN_FULLSCAN     (1<<1)
-#define VERBOSE_FULLSCAN  (1<<2)
-#define CREATE_CRC32      (1<<3)
-#define CHECK_CRC32       (1<<4)
-
-
-#define READABLE_IMAGE    0
-#define READABLE_ECC      0
-#define WRITEABLE_IMAGE   (1<<0)
-#define WRITEABLE_ECC     (1<<1)
-
-#define PRINT_MODE        (1<<4)
-#define CREATE_CRC        ((1<<1) | (1<<5))
-
-char *ApplyAutoSuffix(char*, char*);
 void PrepareDeadSector(void);
 
-void OpenImageAndEcc(ImageInfo**, EccInfo**, int);
-void FreeImageInfo(ImageInfo*);
-void FreeEccInfo(EccInfo*);
-
-void ScanImage(ImageInfo*, EccInfo*, int, GtkWidget*);
-
-ImageInfo* XOpenImage(char*, int, FILE**);
-int VerifyVersion(EccHeader*, int fatal);
-void UnlinkImage(GtkWidget*);
 void CreateEcc(void);
 void FixEcc(void);
+void Verify(void);
+
+/***
+ *** bitmap.c
+ ***/
+
+typedef struct _Bitmap
+{  guint32 *bitmap;
+   gint32 size;
+   gint32 words;
+} Bitmap;
+
+Bitmap* CreateBitmap0(int);
+#define GetBit(bm,bit) (bm->bitmap[(bit)>>5] & (1<<((bit)&31))) 
+#define SetBit(bm,bit) bm->bitmap[(bit)>>5] |= (1<<((bit)&31)) 
+#define ClearBit(bm,bit) bm->bitmap[(bit)>>5] &= ~(1<<((bit)&31)) 
+int CountBits(Bitmap*);
+void FreeBitmap(Bitmap*);
 
 /***
  *** closure.c
  ***/
 
 void InitClosure(void);
+void ClearCrcCache(void);
 void FreeClosure(void);
 void ReadDotfile(void);
 
-/***
- *** compare.c
- ***/
-
-#define COMPARE_IMAGE_SEGMENTS 1000
-
-void Compare(void);
-void ResetCompareWindow(void);
-void CreateCompareWindow(GtkWidget*);
-void AddCompareValues(int, gint64, gint64, gint64, gint64);
-
 /***
  *** crc32.c
  ***/
@@ -417,19 +420,84 @@ void RedrawCurve(Curve*, int, int);
 
 void Byteset(char*);
 void Erase(char*);
-void RandomError(char*);
+void RandomError(char*, char*);
 void RandomImage(char*, char*);
 void SendCDB(char*);
 void ShowSector(char*);
+Bitmap* SimulateDefects(gint64);
 void TruncateImage(char*);
 void ZeroUnreadable(void);
 
+/***
+ *** endian.c
+ ***/
+
+guint32 SwapBytes32(guint32);
+guint64 SwapBytes64(guint64);
+void    SwapEccHeaderBytes(EccHeader*);
+
+/***
+ *** file.c
+ ***/
+
+#define READABLE_IMAGE    0
+#define READABLE_ECC      0
+#define WRITEABLE_IMAGE   (1<<0)
+#define WRITEABLE_ECC     (1<<1)
+#define PRINT_MODE        (1<<4)
+#define CREATE_CRC        ((1<<1) | (1<<5))
+
+ImageInfo* OpenImageFile(EccHeader*, int);
+EccInfo* OpenEccFile(int);
+EccInfo* OpenEccFileOld(int);
+void FreeImageInfo(ImageInfo*);
+void FreeEccInfo(EccInfo*);
+
+char *ApplyAutoSuffix(char*, char*);
+int VerifyVersion(EccHeader*, int fatal);
+void UnlinkImage(GtkWidget*);
+
 /***
  *** fix-window.c
  ***/
 
 void CreateFixWindow(GtkWidget*);
 
+/***
+ *** galois.c
+ ***
+ * This is currently the hardcoded GF(2**8).
+ * gint32 gives abundant space for the GF.
+ * Squeezing it down to guint8 won't probably gain much,
+ * so we implement this defensively here.
+ *
+ * Note that some performance critical stuff needs to
+ * be #included from galois-inlines.h
+ */  
+
+#define SYMBOLSIZE 8
+#define FIELDSIZE (1<<SYMBOLSIZE)
+#define FIELDMAX (FIELDSIZE-1)
+#define ALPHA0 FIELDMAX
+
+#define GENERATOR_POLY 0x187    /* 1 + X + X**2 + X**7 + X**8 */
+#define FIRST_ROOT 112          /* same choices as in CCSDS */                 
+#define PRIM_ELEM 11                  
+#define PRIMTH_ROOT 116         /* prim-th root of 1 */           
+
+typedef struct _GaloisTables
+{  gint32 nroots;        /* degree of generator polynomial */
+   gint32 ndata;         /* data bytes per ecc block */
+
+   gint32 *index_of;     /* log */
+   gint32 *alpha_to;     /* inverse log */
+   gint32 *enc_alpha_to; /* inverse log optimized for encoder */
+   gint32 *gpoly;        /* generator polynomial */
+} GaloisTables;
+
+GaloisTables* CreateGaloisTables(int);
+void FreeGaloisTables(GaloisTables*);
+
 /***
  *** help-dialogs.c
  ***/
@@ -440,6 +508,10 @@ void ShowLog();
 void UpdateLog();
 void AboutDialog();
 
+void AboutText(GtkWidget*, char*, ...);
+void AboutLink(GtkWidget*, char*, char*);
+void AboutTextWithLink(GtkWidget*, char*, char*);
+
 /***
  *** icon-factory.c
  ***/
@@ -450,8 +522,9 @@ void CreateIconFactory();
  *** large-io.c
  ***/
 
-LargeFile *LargeOpen(const char*, int, mode_t);
+LargeFile *LargeOpen(char*, int, mode_t);
 int LargeSeek(LargeFile*, gint64);
+int LargeEOF(LargeFile*);
 int LargeRead(LargeFile*, void*, size_t);
 int LargeWrite(LargeFile*, void*, size_t);
 int LargeClose(LargeFile*);
@@ -463,16 +536,17 @@ int LargeUnlink(char*);
  *** main-window.c
  ***/
 
-#define FIRST_CREATE_WINDOW 4
+#define FIRST_CREATE_WINDOW 3
 
 typedef enum 
 {  ACTION_WELCOME,   /* Tab 0; not really an action   */
    ACTION_STOP,      /* ----   does not have a window */
    ACTION_READ,      /* Tab 1 (linear); Tab 2 (adaptive)*/
    ACTION_SCAN,      /* Tab 1 (linear); Tab 2 (adaptive)*/
-   ACTION_COMPARE,   /* Tab 3 */
-   ACTION_CREATE,    /* CREATE and FIX have separate windows assigned */
-   ACTION_FIX        /* for each method. */
+   ACTION_VERIFY,    /* VERIFY, CREATE and FIX have separate windows assigned */
+   ACTION_CREATE,    /* for each method. */
+   ACTION_CREATE_CONT,
+   ACTION_FIX        
 } MajorActions;
 
 void CreateMainWindow(int*, char***);
@@ -492,6 +566,8 @@ void ContinueWithAction(int);
 #define g_realloc(ptr,size) realloc_ext(ptr,size,__FILE__,__LINE__)
 #define g_strdup(str) strdup_ext(str,__FILE__,__LINE__)
 
+#define g_try_malloc(size) try_malloc_ext(size,__FILE__,__LINE__)
+
 #define g_strdup_printf(format,args...) \
         strdup_printf_ext(format,__FILE__,__LINE__, ## args)
 #define g_strdup_vprintf(format,argp) \
@@ -513,6 +589,7 @@ void ContinueWithAction(int);
 
 void*	malloc_ext(int,char*,int);
 void*	realloc_ext(void*, int, char*, int);
+void*	try_malloc_ext(int,char*,int);
 char*	strdup_ext(const char*,char*,int);
 char*	strdup_printf_ext(char*, char*, int, ...);
 char*	strdup_vprintf_ext(char*, va_list, char*, int);
@@ -541,16 +618,23 @@ GtkWidget* CreateToolBar(GtkWidget*);
 typedef struct _Method
 {  char name[4];                     /* Method name tag */
    char *description;                /* Fulltext description */
+   char *menuEntry;                  /* Text for use in preferences menu */
    void (*create)(struct _Method*);  /* Creates an error correction file */
    void (*fix)(struct _Method*);     /* Fixes a damaged image */
+   void (*verify)(struct _Method*);  /* Verifies image with ecc data */
+   void (*createVerifyWindow)(struct _Method*, GtkWidget*);
    void (*createCreateWindow)(struct _Method*, GtkWidget*);
    void (*createFixWindow)(struct _Method*, GtkWidget*);
    void (*createPrefsPage)(struct _Method*, GtkWidget*);
+   void (*resetVerifyWindow)(struct _Method*);
    void (*resetCreateWindow)(struct _Method*);
    void (*resetFixWindow)(struct _Method*);
+   void (*resetPrefsPage)(struct _Method*);
+   void (*readPreferences)(struct _Method*);
    void (*destroy)(struct _Method*);
    int  tabWindowIndex;              /* our position in the (invisible) notebook */
    void *widgetList;                 /* linkage to window system */
+   EccHeader *lastEh;                /* copy of EccHeader from last EccFileMethod() call */
 } Method;
 
 void BindMethods(void);        /* created by configure in method-link.c */
@@ -559,33 +643,27 @@ void CollectMethods(void);
 void RegisterMethod(Method*);
 void ListMethods(void);
 Method* FindMethod(char*);
+EccHeader* FindHeaderInImage(char*);
+Method *EccFileMethod(int);
 void CallMethodDestructors(void);
 
 /***
  *** misc.c 
  ***/
 
-#ifdef SYS_MINGW
-#define stat _stati64
-#define lseek _lseeki64
-
-int fseeko(FILE*, gint64, int);
-
-#define random rand
-#define srandom(seed) srand(seed)
-#endif
-
 char* sgettext(char*);
 char* sgettext_utf8(char*);
 
 gint64 uchar_to_gint64(unsigned char*);
 void gint64_to_uchar(unsigned char*, gint64);
 
-void AsciiDigest(char*, unsigned char*);
 void CalcSectors(gint64, gint64*, int*);
 
 void PrintCLI(char*, ...);
 void PrintLog(char*, ...);
+void PrintLogFile(char*, ...);
+void Verbose(char*, ...);
+void PrintTimeToLog(GTimer*, char*, ...);
 void PrintProgress(char*, ...);
 void PrintCLIorLabel(GtkLabel*, char*, ...);
 
@@ -617,6 +695,7 @@ void TimedInsensitive(GtkWidget*, int);
  ***/
 
 void CreatePreferencesWindow(void);
+void UpdateMethodPreferences(void);
 void HidePreferences(void);
 void FreePreferences(void*);
 
@@ -641,8 +720,8 @@ void ReadMediumLinear(gpointer);
 void ResetLinearReadWindow();
 void CreateLinearReadWindow(GtkWidget*);
 
-void SetCurveMaxValues(int, gint64, gint64, gint64);
-void AddCurveValues(int, double, gint64, int);
+void InitializeCurve(int, gint64, gint64, gint64);
+void AddCurveValues(int, double, int);
 
 /***
  *** random.c
@@ -690,6 +769,7 @@ void ListAspiDrives(void);
 
 char* DefaultDevice(void);
 gint64 CurrentImageSize(void);
+gint64 CurrentImageCapacity(void);
 
 int SendReadCDB(char*, unsigned char*, int, int);
 
