@@ -1,5 +1,5 @@
 /*  dvdisaster: Additional error correction for optical media.
- *  Copyright (C) 2004,2005 Carsten Gnoerlich.
+ *  Copyright (C) 2004-2006 Carsten Gnoerlich.
  *  Project home page: http://www.dvdisaster.com
  *  Email: carsten@dvdisaster.com  -or-  cgnoerlich@fsfe.org
  *
@@ -43,12 +43,16 @@ typedef struct
    char *formatLinear;
    char *formatAdaptive;
 
+  GtkWidget *radioDrive, *radioISO, *radioECC;
    GtkWidget *radioLinear, *radioAdaptive;
    GtkWidget *rangeToggle, *rangeLab;
    GtkWidget *rangeSpin1, *rangeSpin2;
    GtkWidget *jumpScale;
    GtkWidget *byteEntry, *byteCheck;
    GtkWidget *readAndCreateButton;
+   GtkWidget *mainNotebook;
+   GtkWidget *methodChooser;
+   GtkWidget *methodNotebook;
 
    non_linear_info *jumpScaleInfo;
 } prefs_context;
@@ -75,6 +79,10 @@ static gboolean delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
 
 void HidePreferences(void)
 {  prefs_context *pc = (prefs_context*)Closure->prefsContext;
+   Method *method;
+   int method_index;
+
+   /* Get reading range values */
 
    if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pc->rangeToggle)))   
    {     
@@ -83,6 +91,8 @@ void HidePreferences(void)
    }
    else Closure->readStart = Closure->readEnd = 0;
 
+   /* Get fill byte and recalculate the dead sector marker */
+
    if(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pc->byteCheck)))   
    {  const char *value = gtk_entry_get_text(GTK_ENTRY(pc->byteEntry));
 
@@ -96,6 +106,16 @@ void HidePreferences(void)
    }
    PrepareDeadSector();
 
+   /* Ask currently selected method to update its settings
+      from the preferences */
+
+   method_index = gtk_notebook_get_current_page(GTK_NOTEBOOK(pc->methodNotebook));
+   method = g_ptr_array_index(Closure->methodList, method_index);
+   if(method->readPreferences)
+     method->readPreferences(method);
+
+   /* hide preferences and finish */
+
    gtk_widget_hide(GTK_WIDGET(Closure->prefsWindow));
 }
 
@@ -109,8 +129,7 @@ static void close_cb(GtkWidget *widget, gpointer data)
  */
 
 enum 
-{  TOGGLE_UDF,
-   TOGGLE_READ_CREATE,
+{  TOGGLE_READ_CREATE,
    TOGGLE_UNLINK,
    TOGGLE_SUFFIX,
    TOGGLE_DAO,
@@ -148,11 +167,7 @@ static void toggle_cb(GtkWidget *widget, gpointer data)
    prefs_context *pc = (prefs_context*)Closure->prefsContext;
 
    switch(action)
-   {  case TOGGLE_UDF:
-	Closure->parseUDF = state;
-	break;
-
-      case TOGGLE_READ_CREATE:
+   {  case TOGGLE_READ_CREATE:
 	Closure->readAndCreate = state;
 	if(state && Closure->adaptiveRead)  /* set reading strategy to linear */
 	{  prefs_context *pc = Closure->prefsContext;
@@ -290,6 +305,21 @@ static GtkWidget* non_linear_scale(non_linear_info **nli_ptr, int action, int *v
    return scale;
 }
 
+/*
+ * Image size query method selection 
+ */
+
+static void imgsize_cb(GtkWidget *widget, gpointer data)
+{  int state  = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
+   prefs_context *pc = (prefs_context*)data;
+
+   if(state == TRUE)
+   {  if(pc->radioDrive == widget) Closure->querySize = 0;
+      if(pc->radioISO   == widget) Closure->querySize = 1;
+      if(pc->radioECC   == widget) Closure->querySize = 2;
+   }
+}
+
 /*
  * Read strategy selection 
  */
@@ -371,17 +401,80 @@ static void bytefill_check_cb(GtkWidget *widget, gpointer data)
    }
 }
 
+/***
+ *** Error correction method selection
+ ***/
+
+static void method_select_cb(GtkWidget *widget, gpointer data)
+{  Method *method;
+   prefs_context *pc = (prefs_context*)data;
+   int n;
+
+#if GTK_MINOR_VERSION >= 4
+   n = gtk_combo_box_get_active(GTK_COMBO_BOX(widget));
+#else
+   n = gtk_option_menu_get_history(GTK_OPTION_MENU(widget));
+#endif
+
+   if(n<0 || !pc->methodNotebook)
+     return;
+
+   method = g_ptr_array_index(Closure->methodList, n);
+
+   /* Switch methods if selection changed */
+
+   if(strncmp(Closure->methodName, method->name, 4))
+   {  strncpy(Closure->methodName, method->name, 4);
+      gtk_notebook_set_current_page(GTK_NOTEBOOK(pc->methodNotebook), n);
+   }
+}
+
+/*
+ * Setting the notebook page does not work at creation time.
+ */
+
+static gboolean notebook_idle_func(gpointer data)
+{  prefs_context *pc = (prefs_context*)data;
+   int n;
+
+#if GTK_MINOR_VERSION >= 4
+   n = gtk_combo_box_get_active(GTK_COMBO_BOX(pc->methodChooser));
+#else
+   n = gtk_option_menu_get_history(GTK_OPTION_MENU(pc->methodChooser));
+#endif
+
+   if(n>=0)
+     gtk_notebook_set_current_page(GTK_NOTEBOOK(pc->methodNotebook), n);
+
+   gtk_notebook_set_current_page(GTK_NOTEBOOK(pc->mainNotebook), 0);
+
+   return FALSE;
+}
+
 /***
  *** Assemble and open the preferences window.
  ***/
 
+void UpdateMethodPreferences(void)
+{  int i;
+   
+   for(i=0; i<Closure->methodList->len; i++)
+   {  Method *method = g_ptr_array_index(Closure->methodList, i);
+
+      method->resetPrefsPage(method);
+   }
+}
+
 void CreatePreferencesWindow(void)
 {  
    if(!Closure->prefsWindow)  /* No window to reuse? */
    {  GtkWidget *window, *outer_box, *hbox, *vbox, *vbox2, *notebook, *space, *button, *scale, *frame;
-      GtkWidget *lab, *spin, *radio1, *radio2, *check, *entry;
+     GtkWidget *lab, *spin, *radio1, *radio2, *radio3, *check, *entry;
+#if GTK_MINOR_VERSION < 4
+      GtkWidget *option_menu_strip;
+#endif
       prefs_context *pc = g_malloc0(sizeof(prefs_context));
-      Method *method;
+      int i, method_idx = 0;
  
       Closure->prefsContext = pc;
  
@@ -402,7 +495,7 @@ void CreatePreferencesWindow(void)
       outer_box = gtk_vbox_new(FALSE, 0);
       gtk_container_add(GTK_CONTAINER(window), outer_box);
 
-      notebook = gtk_notebook_new();
+      notebook = pc->mainNotebook = gtk_notebook_new();
       gtk_box_pack_start(GTK_BOX(outer_box), notebook, TRUE, TRUE, 0);
 
       space = gtk_image_new();
@@ -424,22 +517,63 @@ void CreatePreferencesWindow(void)
       frame = gtk_frame_new(_utf("Medium and image filesystem"));
       gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
 
-      button = gtk_check_button_new_with_label(_utf("Use information from ISO/UDF filesystem (experimental)"));
-      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), Closure->parseUDF);
-      g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggle_cb), GINT_TO_POINTER(TOGGLE_UDF));
-      gtk_container_add(GTK_CONTAINER(frame), button);
-      gtk_container_set_border_width(GTK_CONTAINER(button), 10);
+      vbox2 = gtk_vbox_new(FALSE, 15);
+      gtk_container_set_border_width(GTK_CONTAINER(vbox2), 10);
+      gtk_container_add(GTK_CONTAINER(frame), vbox2);
+
+      hbox = gtk_hbox_new(FALSE, 4);
+
+      lab = gtk_label_new(_utf("Get Image size from: ")); 
+      gtk_box_pack_start(GTK_BOX(hbox), lab, FALSE, FALSE, 0);
+
+      radio1 = pc->radioDrive = gtk_radio_button_new(NULL);
+      g_signal_connect(G_OBJECT(radio1), "toggled", G_CALLBACK(imgsize_cb), pc);
+      gtk_box_pack_start(GTK_BOX(hbox), radio1, FALSE, FALSE, 0);
+      lab = gtk_label_new(_utf("Drive"));
+      gtk_container_add(GTK_CONTAINER(radio1), lab);
+
+      radio2 = pc->radioISO = gtk_radio_button_new_from_widget(GTK_RADIO_BUTTON(radio1));
+      g_signal_connect(G_OBJECT(radio2), "toggled", G_CALLBACK(imgsize_cb), pc);
+      gtk_box_pack_start(GTK_BOX(hbox), radio2, FALSE, FALSE, 0);
+      lab = gtk_label_new(_utf("ISO/UDF"));
+      gtk_container_add(GTK_CONTAINER(radio2), lab);
+
+      radio3 = pc->radioECC = gtk_radio_button_new_from_widget(GTK_RADIO_BUTTON(radio2));
+      g_signal_connect(G_OBJECT(radio3), "toggled", G_CALLBACK(imgsize_cb), pc);
+      gtk_box_pack_start(GTK_BOX(hbox), radio3, FALSE, FALSE, 0);
+      lab = gtk_label_new(_utf("ECC/RS02"));
+      gtk_container_add(GTK_CONTAINER(radio3), lab);
+
+      switch(Closure->querySize)
+      {  case 0: gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio1), TRUE); break;
+         case 1: gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio2), TRUE); break;
+         case 2: gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radio3), TRUE); break;
+      }
+
+      gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
+
+
 
       /* file extension */
 
-      frame = gtk_frame_new(_utf("File name suffixes"));
+      frame = gtk_frame_new(_utf("Local files (on hard disc)"));
       gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
 
-      button = gtk_check_button_new_with_label(_utf("automatically add .img and .ecc file suffixes"));
+      vbox2 = gtk_vbox_new(FALSE, 15);
+      gtk_container_set_border_width(GTK_CONTAINER(vbox2), 10);
+      gtk_container_add(GTK_CONTAINER(frame), vbox2);
+
+      button = gtk_check_button_new_with_label(_utf("Automatically add .iso and .ecc file suffixes"));
       gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), Closure->autoSuffix);
       g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggle_cb), GINT_TO_POINTER(TOGGLE_SUFFIX));
-      gtk_container_add(GTK_CONTAINER(frame), button);
-      gtk_container_set_border_width(GTK_CONTAINER(button), 10);
+      gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
+
+      /* 2GB button */
+
+      button = gtk_check_button_new_with_label(_utf("Split files into segments <= 2GB"));
+      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), Closure->splitFiles);
+      g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggle_cb), GINT_TO_POINTER(TOGGLE_2GB));
+      gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
 
       /*** Automatic file creation and deletion */
 
@@ -556,13 +690,6 @@ void CreatePreferencesWindow(void)
       g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggle_cb), GINT_TO_POINTER(TOGGLE_DAO));
       gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
 
-      /* 2GB button */
-
-      button = gtk_check_button_new_with_label(_utf("Split image into files <= 2GB"));
-      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), Closure->splitFiles);
-      g_signal_connect(G_OBJECT(button), "toggled", G_CALLBACK(toggle_cb), GINT_TO_POINTER(TOGGLE_2GB));
-      gtk_box_pack_start(GTK_BOX(vbox2), button, FALSE, FALSE, 0);
-
       /* byte filling */
       
       hbox = gtk_hbox_new(FALSE, 4);
@@ -613,12 +740,81 @@ void CreatePreferencesWindow(void)
 
       /*** "Error correction" page */
 
+      /* Method chooser menu */
+
       vbox = create_page(notebook, _utf("Error correction"));
 
-      method = FindMethod(Closure->methodName); 
-      if(method) method->createPrefsPage(method, vbox);
+      hbox = gtk_hbox_new(FALSE, 4);
+      lab = gtk_label_new(_utf("Storage method:")); 
+      gtk_box_pack_start(GTK_BOX(hbox), lab, FALSE, FALSE, 0);
+      gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
 
-      gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), 1);
+#if GTK_MINOR_VERSION >= 4
+      pc->methodChooser = gtk_combo_box_new_text();
+
+      g_signal_connect(G_OBJECT(pc->methodChooser), "changed", G_CALLBACK(method_select_cb), pc);
+
+      for(i=0; i<Closure->methodList->len; i++)
+      {  Method *method = g_ptr_array_index(Closure->methodList, i);
+
+	 gtk_combo_box_append_text(GTK_COMBO_BOX(pc->methodChooser), method->menuEntry); 
+
+	 if(!strncmp(Closure->methodName, method->name, 4))
+	   method_idx = i;
+      }
+
+
+      gtk_combo_box_set_active(GTK_COMBO_BOX(pc->methodChooser), method_idx);
+      gtk_box_pack_start(GTK_BOX(hbox), pc->methodChooser, FALSE, FALSE, 0);
+#else
+      pc->methodChooser = gtk_option_menu_new();
+
+      g_signal_connect(G_OBJECT(pc->methodChooser), "changed", G_CALLBACK(method_select_cb), pc);
+      option_menu_strip = gtk_menu_new(); 
+
+      for(i=0; i<Closure->methodList->len; i++)
+      {  Method *method = g_ptr_array_index(Closure->methodList, i);
+	 GtkWidget *item;
+
+	 item = gtk_menu_item_new_with_label(method->menuEntry);
+	 gtk_menu_shell_append(GTK_MENU_SHELL(option_menu_strip), item);
+
+	 if(!strncmp(Closure->methodName, method->name, 4))
+	   method_idx = i;
+      }
+
+      gtk_option_menu_set_menu(GTK_OPTION_MENU(pc->methodChooser), option_menu_strip);
+      gtk_option_menu_set_history(GTK_OPTION_MENU(pc->methodChooser), method_idx);
+      gtk_box_pack_start(GTK_BOX(hbox), pc->methodChooser, FALSE, FALSE, 0);
+#endif
+
+      /* sub pages for individual method configuration */
+      
+      pc->methodNotebook = gtk_notebook_new();
+      gtk_notebook_set_show_tabs(GTK_NOTEBOOK(pc->methodNotebook), FALSE);
+      gtk_notebook_set_show_border(GTK_NOTEBOOK(pc->methodNotebook), FALSE);
+      gtk_box_pack_start(GTK_BOX(vbox), pc->methodNotebook, TRUE, TRUE, 0);
+
+      for(i=0; i<Closure->methodList->len; i++)
+      {  Method *method = g_ptr_array_index(Closure->methodList, i);
+	 GtkWidget *vbox2 = gtk_vbox_new(FALSE, 0);
+	 GtkWidget *ignore = gtk_label_new("method_tab");
+
+	 if(method->createPrefsPage)
+	    method->createPrefsPage(method, vbox2);
+	 else
+	 {  GtkWidget *lab;
+
+	    lab = gtk_label_new("This method has no configuration options.");
+	    gtk_box_pack_start(GTK_BOX(vbox2), lab, TRUE, TRUE, 0);
+	 }
+
+	 gtk_notebook_append_page(GTK_NOTEBOOK(pc->methodNotebook), vbox2, ignore);
+	 if(i==method_idx)
+      gtk_notebook_set_current_page(GTK_NOTEBOOK(pc->methodNotebook), method_idx);
+      }
+
+      g_idle_add(notebook_idle_func, pc);
    }
 
    /* Show the created / reused window */
