@@ -1,5 +1,5 @@
 /*  dvdisaster: Additional error correction for optical media.
- *  Copyright (C) 2004,2005 Carsten Gnoerlich.
+ *  Copyright (C) 2004-2006 Carsten Gnoerlich.
  *  Project home page: http://www.dvdisaster.com
  *  Email: carsten@dvdisaster.com  -or-  cgnoerlich@fsfe.org
  *
@@ -26,55 +26,70 @@
 //#define DEBUG 1
 
 /***
- *** Local data package used during reading
+ *** Local data package used during reading 
  ***/
 
 typedef struct
 {  LargeFile *image;
-   DeviceHandle *dh;
+   struct _DeviceHandle *dh;
+   EccInfo *ei;
+   char *msg;
    unsigned char *bufbase;
-   GTimer *timer;
+   GTimer *speedTimer,*readTimer;
    int unreportedError;
    int earlyTermination;
    int scanMode;
-   gint64 readOK, readErrors;
+   gint64 readOK;
 } read_closure;
 
 static void cleanup(gpointer data)
 {  read_closure *rc = (read_closure*)data;
-   int read_errors = rc->readErrors;
-   int aborted     = rc->earlyTermination;
-   int scan_mode   = rc->scanMode;
+   int full_read = FALSE;
+   int aborted   = rc->earlyTermination;
+   int scan_mode = rc->scanMode;
+
+   if(rc->dh)
+     full_read = (rc->readOK == rc->dh->sectors && !Closure->crcErrors);
 
    Closure->cleanupProc = NULL;
 
    if(Closure->guiMode)
    {  if(rc->unreportedError)
          SwitchAndSetFootline(Closure->readLinearNotebook, 1, Closure->readLinearFootline, 
-			      _("<span color=\"red\">Aborted by unrecoverable error.</span> %lld sectors read, %lld sectors unreadable/skipped so far."),rc->readOK,rc->readErrors); 
+			      _("<span color=\"red\">Aborted by unrecoverable error.</span> %lld sectors read, %lld sectors unreadable/skipped so far."),rc->readOK,Closure->readErrors); 
    }
 
    if(rc->image)   
      if(!LargeClose(rc->image))
        Stop(_("Error closing image file:\n%s"), strerror(errno));
    if(rc->dh)      CloseDevice(rc->dh);
+   if(rc->ei)      FreeEccInfo(rc->ei);
+   if(rc->msg)     g_free(rc->msg);
    if(rc->bufbase) g_free(rc->bufbase);
-   if(rc->timer)   g_timer_destroy(rc->timer);
+   if(rc->speedTimer) g_timer_destroy(rc->speedTimer);
+   if(rc->readTimer)  g_timer_destroy(rc->readTimer);
    g_free(rc);
 
-   if(Closure->readAndCreate && !scan_mode && !aborted)
-   {  if(read_errors)
+   if(Closure->readAndCreate && Closure->guiMode && !strncmp(Closure->methodName, "RS01", 4)
+      && !scan_mode && !aborted)
+   {  if(!full_read)
       {  ModalDialog(GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, NULL,
-		     _("Automatic error correction file creation stopped\n"
-		       "because of read errors in the current medium.\n"));
+		     _("Automatic error correction file creation\n"
+		       "is only possible after a full reading pass.\n"));
 	 AllowActions(TRUE);
       }
-      else ContinueWithAction(ACTION_CREATE); 
+      else ContinueWithAction(ACTION_CREATE_CONT); 
    }
    else 
      if(Closure->guiMode)
        AllowActions(TRUE);
 
+   if(!full_read && Closure->crcCache)
+     ClearCrcCache();
+
+   if(scan_mode)   /* we haven't created an image, so throw away the crc sums */
+     ClearCrcCache();
+
    g_thread_exit(0);
 }
 
@@ -96,11 +111,12 @@ void ReadMediumLinear(gpointer data)
    int nsectors; 
    gint64 start,end,s,read_marker;
    gint64 image_size;
-   int last_percent,percent;
+   int percent, last_percent;
    double last_speed = -1.0;
    int first_speed_value = TRUE;
    int rereading = 0;
-   gint previous_errors = 0;
+   gint previous_read_errors = 0;
+   gint previous_crc_errors = 0;
    unsigned char *buf = NULL;
    char *t;
    int status,n;
@@ -108,6 +124,9 @@ void ReadMediumLinear(gpointer data)
    double speed = 0.0;
    int tao_tail = 0;
    int ignore_fatal = FALSE;
+   struct MD5Context md5ctxt;
+
+   Closure->additionalSpiralColor = -1;
 
    /*** Register the cleanup procedure so that Stop() can abort us properly. */
 
@@ -131,26 +150,29 @@ void ReadMediumLinear(gpointer data)
    }
 
    if(Closure->readAndCreate && !scan_mode)
-   {  struct stat mystat;
-      if(stat(Closure->eccName, &mystat) != -1)
-      {  int answer = ModalDialog(GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL, NULL,
-				  _("Automatic error correction file creation is enabled,\n"
-				    "and \"%s\" already exists.\n"
-				    "Overwrite it?\n"),
-				  Closure->eccName);
+   {  gint64 ignore;
+      if(LargeStat(Closure->eccName, &ignore))
+      { if(Closure->guiMode)
+	{  int answer = ModalDialog(GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL, NULL,
+				    _("Automatic error correction file creation is enabled,\n"
+				      "and \"%s\" already exists.\n"
+				      "Overwrite it?\n"),
+				    Closure->eccName);
 
-	if(!answer)
-	{  SwitchAndSetFootline(Closure->readLinearNotebook, 1, Closure->readLinearFootline, 
+	  if(!answer)
+	  {  SwitchAndSetFootline(Closure->readLinearNotebook, 1, Closure->readLinearFootline, 
 				_("<span color=\"red\">Aborted by user request!</span>")); 
-	   rc->unreportedError = FALSE;
-	   goto terminate;
+	     rc->unreportedError = FALSE;
+	     goto terminate;
+	  }
 	}
       }
    }
 
    /*** Timer setup */
 
-   rc->timer = g_timer_new();
+   rc->speedTimer = g_timer_new();
+   rc->readTimer  = g_timer_new();
 
    /*** Align the buffer at a 4096 boundary.
 	Might be needed by some SCSI drivers. */
@@ -163,7 +185,24 @@ void ReadMediumLinear(gpointer data)
 
    rc->dh = OpenAndQueryDevice(Closure->device);
    sectors = rc->dh->sectors;
-   rc->readErrors = rc->readOK = 0;
+   Closure->readErrors = Closure->crcErrors = rc->readOK = 0;
+
+   /*** See if the current ecc file belongs to the medium */
+
+   rc->ei = OpenEccFile(READABLE_ECC | PRINT_MODE);
+   if(rc->ei) /* Compare the fingerprint sectors */
+   {  struct MD5Context md5ctxt;
+      guint8 fingerprint[16];
+
+      status = ReadSectors(rc->dh, buf, rc->ei->eh->fpSector, 1);
+
+      MD5Init(&md5ctxt);
+      MD5Update(&md5ctxt, buf, 2048);
+      MD5Final(fingerprint, &md5ctxt);
+
+      if(!status && !memcmp(fingerprint, rc->ei->eh->mediumFP, 16))
+	Closure->checkCrc = TRUE;
+   }
 
    /*** See if user wants to limit the read range. */
 
@@ -194,20 +233,28 @@ void ReadMediumLinear(gpointer data)
    /*** See if we already have an image file. */
 
    if(scan_mode)  /* don't care if in scan mode */
-   {  char *t = _("Scanning medium for read errors.");
+   {  
+      rc->msg = g_strdup(_("Scanning medium for read errors."));
 
-      PrintLog("%s\n",t);
+      PrintLog("%s\n", rc->msg);
       if(Closure->guiMode)
-         SetLabelText(GTK_LABEL(Closure->readLinearHeadline), 
-		      "<big>%s</big>\n<i>%s</i>",t,rc->dh->mediumDescr);
+      {  if(Closure->checkCrc)
+	   SetLabelText(GTK_LABEL(Closure->readLinearHeadline), 
+			"<big>%s</big>\n<i>- %s -</i>", rc->msg,
+			_("Reading CRC information from ecc file"));
 
+         else
+	   SetLabelText(GTK_LABEL(Closure->readLinearHeadline), 
+			"<big>%s</big>\n<i>%s</i>", rc->msg, rc->dh->mediumDescr);
+      }
       read_marker = 0;
    } 
    else          
    { 
 reopen_image:
      if(!LargeStat(Closure->imageName, &image_size))
-     {  char *t = _("Reading new medium image.");
+     {  
+        rc->msg = g_strdup(_("Reading new medium image."));
 
         /* Start with a fresh file. */
 
@@ -216,14 +263,20 @@ reopen_image:
 
         PrintLog(_("Creating new %s image.\n"),Closure->imageName);
 	if(Closure->guiMode)
-	  SetLabelText(GTK_LABEL(Closure->readLinearHeadline),
-		       "<big>%s</big>\n<i>%s</i>",t,rc->dh->mediumDescr);
+	{  if(Closure->checkCrc)
+	     SetLabelText(GTK_LABEL(Closure->readLinearHeadline),
+			  "<big>%s</big>\n<i>%s</i>", rc->msg,
+			  _("Reading CRC information from ecc file"));
+	   else
+	     SetLabelText(GTK_LABEL(Closure->readLinearHeadline),
+			  "<big>%s</big>\n<i>%s</i>", rc->msg, rc->dh->mediumDescr);
+	}
         rereading = 0;
 	read_marker = 0;
      }
      else
      {  char *t = _("Completing existing medium image.");
-        int unknown_footprint = FALSE;
+        int unknown_fingerprint = FALSE;
 
         /* Use the existing file as a starting point.
 	   Set the read_marker at the end of the file
@@ -236,19 +289,19 @@ reopen_image:
         rereading = 1;
 	read_marker = image_size / 2048;
 
-	/* See if the media and image footprints match */
+	/* See if the media and image fingerprints match */
 
-	if(!LargeSeek(rc->image, (gint64)(2048*FOOTPRINT_SECTOR)))
-	  unknown_footprint = TRUE;
+	if(!LargeSeek(rc->image, (gint64)(2048*FINGERPRINT_SECTOR)))
+	  unknown_fingerprint = TRUE;
 	else
 	{  n = LargeRead(rc->image, buf, 2048);
-	   status = ReadSectors(rc->dh, buf+2048, FOOTPRINT_SECTOR, 1);
+	   status = ReadSectors(rc->dh, buf+2048, FINGERPRINT_SECTOR, 1);
 
 	   if(n != 2048 || status || !memcmp(buf, Closure->deadSector, 2048))
-	     unknown_footprint = TRUE;
+	     unknown_fingerprint = TRUE;
         }
 
-	if(!unknown_footprint && memcmp(buf, buf+2048, 2048))
+	if(!unknown_fingerprint && memcmp(buf, buf+2048, 2048))
 	{  	  
 	   if(!Closure->guiMode)
 	     Stop(_("Image file does not match the CD/DVD."));
@@ -274,12 +327,19 @@ reopen_image:
 	/* If the image is not complete yet, first aim to read the
 	   unvisited sectors before trying to re-read the missing ones. */
 
+	Closure->checkCrc = 0; /* makes only sense if image is completely read */
+
 	if(!Closure->readStart && !Closure->readEnd && read_marker < sectors-1)
 	{  PrintLog(_("Completing image %s. Continuing with sector %lld.\n"),
 	            Closure->imageName, read_marker);
 	   start = read_marker;
+	   Closure->additionalSpiralColor = 0;  /* blue */
 	}
-	else PrintLog(_("Completing image %s. Only missing sectors will be read.\n"), Closure->imageName);
+	else 
+	{  PrintLog(_("Completing image %s. Only missing sectors will be read.\n"), Closure->imageName);
+	   Closure->additionalSpiralColor = 3;  /* dark green*/
+	}
+	   
 	if(Closure->guiMode)
 	  SetLabelText(GTK_LABEL(Closure->readLinearHeadline),
 		       "<big>%s</big>\n<i>%s</i>",t,rc->dh->mediumDescr);
@@ -287,7 +347,7 @@ reopen_image:
    }
 
    if(Closure->guiMode)
-     SetCurveMaxValues(rc->dh->maxRate, start, end, sectors);
+     InitializeCurve(rc->dh->maxRate, start, end, sectors);
 
    /*** If start > read_marker, fill the gap with dead sector markers. */
 
@@ -295,12 +355,71 @@ reopen_image:
    {  s = read_marker;
 
       if(!LargeSeek(rc->image, (gint64)(2048*s)))
-	Stop(_("lseek/fill failed for %s, sector %lld"),Closure->imageName,s);
+	Stop(_("Failed seeking to sector %lld in image [%s]: %s"),
+	     s, "fill", strerror(errno));
 
       while(s++ < start)
       {  n = LargeWrite(rc->image, Closure->deadSector, 2048);
 	 if(n != 2048)
-	   Stop(_("Write error filling %s:\n%s"),Closure->imageName,strerror(errno));
+	   Stop(_("Failed writing to sector %lld in image [%s]: %s"),
+		s, "fill", strerror(errno));
+      }
+   }
+
+   /*** Memory for the CRC32 sums is needed in two cases: */
+
+   /* a) a full image read is attempted, and the image CRC32 
+         and md5sum are calculated on the fly. */
+
+   if(   !scan_mode && !rereading && start == 0 && end == sectors-1)
+   {  Closure->crcCache = g_try_malloc(sizeof(guint32) * sectors);
+
+      if(Closure->crcCache)
+	Closure->crcImageName = g_strdup(Closure->imageName);
+
+      MD5Init(&md5ctxt);
+   }
+
+   /* b) we have a suitable ecc file and compare CRC32sum against
+         it while reading */
+
+   if(Closure->checkCrc)
+   {  gint64 crc_sectors = rc->ei->sectors;
+
+      if(sectors > crc_sectors)  /* Allows completion of crc info */
+	crc_sectors = sectors;   /* when image contains additional sectors */
+
+      Closure->crcCache = g_try_malloc(sizeof(guint32) * crc_sectors);
+
+      if(!Closure->crcCache)
+      {  Closure->checkCrc = FALSE;
+      }
+      else
+      {  guint32 *cache = Closure->crcCache;
+         int i=0;
+
+	 if(!LargeSeek(rc->ei->file, (gint64)sizeof(EccHeader)))
+	   Stop(_("Failed skipping the ecc header: %s"),strerror(errno));
+
+	 PrintCLI("%s ...",_("Reading CRC information from ecc file"));
+
+	 while(i<crc_sectors)
+	 {  int n = i+512<crc_sectors ? 512 : crc_sectors - i;
+
+	    if(LargeRead(rc->ei->file, cache, 4*n) != 4*n)
+	      Stop(_("Error reading CRC information: %s"), strerror(errno));
+
+	    cache += n;
+	    i+=n;
+	 }
+
+         PrintCLI(_("done.\n"));
+
+	 if(Closure->guiMode)
+	   SetLabelText(GTK_LABEL(Closure->readLinearHeadline),
+			"<big>%s</big>\n<i>%s</i>", rc->msg, rc->dh->mediumDescr);
+
+	 MD5Init(&md5ctxt);
       }
    }
 
@@ -316,20 +435,21 @@ reopen_image:
      SwitchAndSetFootline(Closure->readLinearNotebook, 0, Closure->readLinearFootline, "ignore");
 
    ReadSectors(rc->dh, buf, start, 1); /* eliminate initial seek time from timing */
-   g_timer_start(rc->timer);
+   g_timer_start(rc->speedTimer);
+   g_timer_start(rc->readTimer);
 
    /*** Read the medium image. */
 
-   last_percent = 0;
-   last_read_ok = last_errors_printed = 0;
    s = start;
+   last_percent = (1000*s)/sectors;
+   last_read_ok = last_errors_printed = 0;
 
    while(s<=end)
    {  if(Closure->stopActions)   /* somebody hit the Stop button */
       {
 	SwitchAndSetFootline(Closure->readLinearNotebook, 1, Closure->readLinearFootline, 
 			     _("<span color=\"red\">Aborted by user request!</span> %lld sectors read, %lld sectors unreadable/skipped so far."),
-			     rc->readOK,rc->readErrors); 
+			     rc->readOK,Closure->readErrors); 
 	rc->unreportedError = FALSE;  /* suppress respective error message */
         goto terminate;
       }
@@ -360,7 +480,8 @@ reread:
 	 int num_compare = nsectors;
 
 	 if(!LargeSeek(rc->image, (gint64)(2048*s)))
-	   Stop(_("lseek/1 failed for %s, sector %lld"),Closure->imageName,s);
+	   Stop(_("Failed seeking to sector %lld in image [%s]: %s"),
+		s, "reread", strerror(errno));
 
 	 if(s+nsectors > read_marker)
 	   num_compare = read_marker-s;
@@ -423,7 +544,7 @@ printf("Sector %lld: reading %2d sectors\n",s,nsectors);
 	 if(!answer)
 	 {  SwitchAndSetFootline(Closure->readLinearNotebook, 1, Closure->readLinearFootline, 
 				_("<span color=\"red\">Aborted by user request!</span> %lld sectors read, %lld sectors unreadable/skipped so far."),
-				 rc->readOK,rc->readErrors); 
+				 rc->readOK,Closure->readErrors); 
 	    rc->unreportedError = FALSE;  /* suppress respective error message */
 	    goto terminate;
 	 }
@@ -434,6 +555,11 @@ printf("Sector %lld: reading %2d sectors\n",s,nsectors);
       if(status)
       {  int nfill;
 
+	 /* Disable on the fly checksum calculation */
+
+	 if(Closure->crcCache && !Closure->checkCrc)
+	   ClearCrcCache();
+
 	 /* Determine number of sectors to skip forward. 
 	    Make sure not to skip past the media end
 	    and to land at a multiple of 16. */
@@ -456,12 +582,14 @@ printf("Sector %lld: reading %2d sectors\n",s,nsectors);
 printf("Sector %lld: filling %d sectors\n",s,nfill);
 #endif
 	    if(!LargeSeek(rc->image, (gint64)(2048*s)))
-	      Stop(_("lseek/3 failed for %s, sector %lld"),Closure->imageName,s);
+	      Stop(_("Failed seeking to sector %lld in image [%s]: %s"),
+		   s, "nfill", strerror(errno));
 
 	    for(i=0; i<nfill; i++)
 	    {  n = LargeWrite(rc->image, Closure->deadSector, 2048);
 	       if(n != 2048)
-		 Stop(_("Write error on %s:\n%s"),Closure->imageName,strerror(errno));
+		 Stop(_("Failed writing to sector %lld in image [%s]: %s"),
+		      s, "nfill", strerror(errno));
 	    }
 	 }
 	 
@@ -476,7 +604,7 @@ printf("Sector %lld: filling %d sectors\n",s,nfill);
 	 {  PrintCLIorLabel(Closure->status,
 			    _("Sector %lld: %s Skipping %d sectors.\n"),
 			    s, GetLastSenseString(FALSE), nfill-1);  
-	    rc->readErrors+=nfill; 
+	    Closure->readErrors+=nfill; 
 	    s+=nfill-nsectors;   /* nsectors will be added again after the goto */
 	    goto step_counter;
 	 }
@@ -496,100 +624,177 @@ printf("Sector %lld: filling %d sectors\n",s,nfill);
 	    {  PrintCLIorLabel(Closure->status,
 			       _("Sector %lld: %s\n"),
 			       s, GetLastSenseString(FALSE));  
-	      if(s >= sectors - 2) tao_tail++;
-	      rc->readErrors++;
+	       if(s >= sectors - 2) tao_tail++;
+	       Closure->readErrors++;
 	    }
 	 }
       }
       else
-      {
+      { int i;
 	/*** Store sector(s) in the image file */
 
 	rc->readOK += nsectors;
 
 	if(!scan_mode)
 	{  if(!LargeSeek(rc->image, (gint64)(2048*s)))
-	     Stop(_("lseek/2 failed for %s, sector %lld: %s"),Closure->imageName,s,strerror(errno));
+	     Stop(_("Failed seeking to sector %lld in image [%s]: %s"),
+		  s, "store", strerror(errno));
 
 #ifdef DEBUG
 printf("Sector %lld: writing %2d sectors\n",s,nsectors);
 #endif
            n = LargeWrite(rc->image, buf, 2048*nsectors);
 	   if(n != 2048*nsectors)
-	     Stop(_("Write error on %s:\n%s"),Closure->imageName,strerror(errno));
+	     Stop(_("Failed writing to sector %lld in image [%s]: %s"),
+		  s, "store", strerror(errno));
+
+	   /* On-the-fly CRC/MD5 calculation */
+
+	   if(!Closure->checkCrc && Closure->crcCache)  
+	   {  for(i=0; i<nsectors; i++)
+		Closure->crcCache[s+i] = Crc32(buf+2048*i, 2048);
+
+	      MD5Update(&md5ctxt, buf, 2048*nsectors);
+	   }
 	}
+
+	/* On-the-fly CRC testing */         
+
+	if(Closure->checkCrc)
+	  for(i=0; i<nsectors; i++)
+	  {  guint32 crc = Crc32(buf+2048*i, 2048);
+
+	     if(s+i < rc->ei->sectors)
+	     {  if(Closure->crcCache[s+i] != crc)
+	        {  PrintCLI(_("* CRC error, sector: %lld\n"), (long long int)s+i);
+		   Closure->crcErrors++;
+		}
+	     }
+	     else Closure->crcCache[s+i] = crc; /* add CRCsums for additional sectors */
+
+	     MD5Update(&md5ctxt, buf+2048*i, 2048);
+	  }
       }
 
       /*** Step the progress counter */
 
 step_counter:
-      if(Closure->guiMode && last_errors_printed != rc->readErrors)
+      if(Closure->guiMode && last_errors_printed != Closure->readErrors)
       {  SetLabelText(GTK_LABEL(Closure->readLinearErrors), 
-		      _("Unreadable / skipped sectors: %lld"), rc->readErrors);
-	 last_errors_printed = rc->readErrors;
+		      _("Unreadable / skipped sectors: %lld"), Closure->readErrors);
+	 last_errors_printed = Closure->readErrors;
       }
 
       s += nsectors;
       if(s>read_marker) read_marker=s;
       percent = (1000*s)/sectors;
+
       if(last_percent != percent) 
       {  gulong ignore;
+	 int color;
 
 	 if(rc->readOK <= last_read_ok)  /* anything read since last sample? */
-	   speed = 0.0;  /* nothing read */
+	 {  speed = 0.0;                 /* nothing read */
+	    if(Closure->readErrors-previous_read_errors > 0)
+	         color = 2;
+	    else if(Closure->crcErrors-previous_crc_errors > 0)
+	         color = 4;
+	    else color = Closure->additionalSpiralColor;
+
+	    if(Closure->guiMode)
+	       AddCurveValues(percent, speed, color);
+	    last_percent    = percent;
+	    last_speed      = speed;
+	    previous_read_errors = Closure->readErrors;
+	    previous_crc_errors  = Closure->crcErrors;
+	    last_read_ok    = rc->readOK;
+	 }
 	 else
 	 {  double kb_read = (rc->readOK - last_read_ok) * 2.0;
-	    double elapsed = g_timer_elapsed(rc->timer, &ignore);
+	    double elapsed = g_timer_elapsed(rc->speedTimer, &ignore);
 	    double kb_sec  = kb_read / elapsed;
 
-	    if(elapsed > 0.1)  /* sample speed for at least 0.1sec */
-	    {  if(first_speed_value) 
-	       {  speed = kb_sec / rc->dh->singleRate; 
-		  if(Closure->guiMode)
-		    AddCurveValues(0, speed, rc->readErrors, rc->readErrors-previous_errors);
-		  first_speed_value = FALSE;
-	       }
-	       else speed = (speed + kb_sec / rc->dh->singleRate) / 2.0;
+	    if(Closure->readErrors-previous_read_errors > 0)
+	         color = 2;
+	    else if(Closure->crcErrors-previous_crc_errors > 0)
+	         color = 4;
+	    else color = 1;
+
+	    if(first_speed_value)
+	    {   speed = kb_sec / rc->dh->singleRate;
+
+	        if(Closure->guiMode)
+		{  AddCurveValues(last_percent, speed, color);
+		   AddCurveValues(percent, speed, color);
+		}
+
+		first_speed_value = FALSE;
+		last_percent      = percent;
+		last_speed        = speed;
+		previous_read_errors = Closure->readErrors;
+		previous_crc_errors  = Closure->crcErrors;
+		last_read_ok      = rc->readOK;
+	    }
+	    else
+	    {  speed = (speed + kb_sec / rc->dh->singleRate) / 2.0;
 	       if(speed>99.9) speed=99.9;
 
-	       last_read_ok = rc->readOK;
-	       g_timer_start(rc->timer);
-	    }
+	       if(Closure->guiMode)
+		 AddCurveValues(percent, speed, color);
+	       if(Closure->speedWarning && last_speed > 0.5)
+	       {  double delta = speed - last_speed;
+		  double abs_delta = fabs(delta);
+		  double sp = (100.0*abs_delta) / last_speed; 
 
-	 }
-
-	 if(Closure->guiMode) 
-	 {  AddCurveValues(percent, speed, rc->readErrors, rc->readErrors-previous_errors);
-	    previous_errors = rc->readErrors;
-	 }
-	 else 
-	 {  if(Closure->speedWarning && last_speed > 0.5)
-	    {  double delta = speed - last_speed;
-	       double abs_delta = fabs(delta);
-	       double sp = (100.0*abs_delta) / last_speed; 
-
-	       if(sp >= Closure->speedWarning)
-	       {  if(delta > 0.0)
-		     PrintCLI(_("Sector %lld: Speed increased to %4.1fx\n"), s, fabs(speed));
-	          else
-		     PrintCLI(_("Sector %lld: Speed dropped to %4.1fx\n"), s, fabs(speed));
+		  if(sp >= Closure->speedWarning)
+		  {  if(delta > 0.0)
+		       PrintCLI(_("Sector %lld: Speed increased to %4.1fx\n"), s, fabs(speed));
+		     else
+		       PrintCLI(_("Sector %lld: Speed dropped to %4.1fx\n"), s, fabs(speed));
+		  }
 	       }
+
+	       PrintProgress(_("Read position: %3d.%1d%% (%4.1fx)"),
+			     percent/10,percent%10,speed);
+
+	       last_percent    = percent;
+	       last_speed      = speed;
+	       previous_read_errors = Closure->readErrors;
+	       previous_crc_errors = Closure->crcErrors;
+	       last_read_ok    = rc->readOK;
+	       g_timer_start(rc->speedTimer);
 	    }
-	    PrintProgress(_("Read position: %3d.%1d%% (%4.1fx)"),
-			  percent/10,percent%10,speed);
 	 }
-	 last_percent = percent;
-	 last_speed = speed;
       }
    }
 
+   if(Closure->crcCache)
+     MD5Final(Closure->md5Cache, &md5ctxt);
+
    if(rereading)
-   {  if(!rc->readErrors) t = g_strdup_printf(_("%lld sectors read.     "),rc->readOK);
-      else                t = g_strdup_printf(_("%lld sectors read; %lld unreadable sectors."),rc->readOK,rc->readErrors);
+   {  if(!Closure->readErrors) t = g_strdup_printf(_("%lld sectors read.     "),rc->readOK);
+      else                t = g_strdup_printf(_("%lld sectors read; %lld unreadable sectors."),rc->readOK,Closure->readErrors);
    }
    else
-   {  if(!rc->readErrors) t = g_strdup_printf(_("All sectors successfully read."));
-      else                t = g_strdup_printf(_("%lld unreadable sectors."),rc->readErrors);
+   {  if(!Closure->readErrors && !Closure->crcErrors) 
+      {  
+	 if(Closure->checkCrc && rc->ei)
+	 {  
+	    if(rc->dh->sectors != rc->ei->sectors)
+	      t = g_strdup_printf(_("All sectors successfully read, but wrong image length (%lld sectors difference)"), rc->dh->sectors - rc->ei->sectors);
+	    else if(memcmp(rc->ei->eh->mediumSum, Closure->md5Cache, 16))
+	      t = g_strdup_printf(_("All sectors successfully read, but wrong image checksum."));
+	    else t = g_strdup_printf(_("All sectors successfully read. Checksums match."));
+	 }
+	 else t = g_strdup_printf(_("All sectors successfully read."));
+      }
+      else 
+      {  if(Closure->readErrors && !Closure->crcErrors)
+	      t = g_strdup_printf(_("%lld unreadable sectors."),Closure->readErrors);
+         else if(!Closure->readErrors && Closure->crcErrors)
+	      t = g_strdup_printf(_("%lld CRC errors."),Closure->crcErrors);
+	 else t = g_strdup_printf(_("%lld CRC errors, %lld unreadable sectors."),Closure->crcErrors, Closure->readErrors);
+      }
    }
    PrintLog("\n%s\n",t);
    if(Closure->guiMode)
@@ -600,21 +805,23 @@ step_counter:
    }
    g_free(t);
 
-   if(rc->dh->subType != DVD && tao_tail && tao_tail == rc->readErrors && !Closure->noTruncate)
+   PrintTimeToLog(rc->readTimer, "for reading/scanning.\n");
+
+   if(rc->dh->subType != DVD && tao_tail && tao_tail == Closure->readErrors && !Closure->noTruncate)
    {  int answer;
    
       if(Closure->guiMode)
         answer = ModalWarning(GTK_MESSAGE_WARNING, GTK_BUTTONS_OK_CANCEL, NULL,
 			      _("%d sectors missing at the end of the disc.\n"
 				"This is okay if the CD was written in TAO (track at once) mode.\n"
-				"The Image will be trimmed accordingly. See the manual for details.\n"),
+				"The Image will be truncated accordingly. See the manual for details.\n"),
 			      tao_tail);
       else 
         answer = ModalWarning(GTK_MESSAGE_WARNING, GTK_BUTTONS_OK_CANCEL, NULL,
 			      _("%d sectors missing at the end of the disc.\n"
 				"This is okay if the CD was written in TAO (track at once) mode.\n"
-				"The Image will be trimmed accordingly. See the manual for details.\n"
-				"Use the --dao option to disable image trimming.\n"),
+				"The Image will be truncated accordingly. See the manual for details.\n"
+				"Use the --dao option to disable image truncating.\n"),
 			      tao_tail);
      
       sectors -= tao_tail;
@@ -623,7 +830,7 @@ step_counter:
         if(!LargeTruncate(rc->image, (gint64)(2048*sectors)))
 	  Stop(_("Could not truncate %s: %s\n"),Closure->imageName,strerror(errno));
    }
-   else if(rc->readErrors) exitCode = EXIT_FAILURE;
+   else if(Closure->readErrors) exitCode = EXIT_FAILURE;
 
    /*** Close and clean up */
 
