@@ -1,5 +1,5 @@
 /*  dvdisaster: Additional error correction for optical media.
- *  Copyright (C) 2004,2005 Carsten Gnoerlich.
+ *  Copyright (C) 2004-2006 Carsten Gnoerlich.
  *  Project home page: http://www.dvdisaster.com
  *  Email: carsten@dvdisaster.com  -or-  cgnoerlich@fsfe.org
  *
@@ -32,6 +32,9 @@ typedef struct _IsoInfo
 
 void FreeIsoInfo(IsoInfo*);
 
+EccHeader* FindHeaderInMedium(DeviceHandle*, gint64);
+gint64 MediumLengthFromRS02(DeviceHandle*, gint64);
+
 int ExamineUDF(DeviceHandle*);
 
 #endif
